"""Helper functions for statistical evalutation of (dataframe-based) data."""

from collections.abc import Hashable

import numpy as np
import pandas as pd


def dist_table(
    df: pd.DataFrame,
    category_cols: str | list[str],
    id_cols: str | list[str] | None = None,
    value_col: str | None = None,
    domains: dict[str, list[Hashable] | np.ndarray | pd.Index] = {},
) -> pd.Series:
    """Return a frequency table of the distribution of unique entities.

    Entities are identified by ``id_cols``. Distribution is presented over
    unique categories in ``category_cols``.

    Args:
        df: Dataframe to evaluate.
        category_cols: Columns to evaluate distribution over.
        id_cols: Columns to identify entities by.
        value_col: Unique values per entity to sum up.
        domains:
            Force the distribution to be evaluated over these domains,
            filling missing values with 0.

    Returns:
        Series of the distribution's values (count or sum) given the categories
        in the index.
    """
    id_cols = (
        [id_cols]
        if isinstance(id_cols, str)
        else id_cols
        if id_cols is not None
        else [n or "index" for n in df.index.names]
    )
    category_cols = [
        *([category_cols] if isinstance(category_cols, str) else category_cols)
    ]

    counts = (
        df.reset_index()[
            [*category_cols, *id_cols, *([value_col] if value_col else [])]
        ]
        .groupby(by=category_cols, group_keys=True)
        .apply(
            lambda df: len(df.drop_duplicates())
            if value_col is None
            else df.drop_duplicates(subset=id_cols)[value_col].sum()
        )
        .rename("value")
    )

    if len(domains) > 0:
        if len(category_cols) == 1:
            col = category_cols[0]
            domain = domains.get(col)

            if domain is not None:
                counts = counts.reindex(
                    domain,
                    fill_value=0,
                ).rename_axis(index=col)
        else:
            count_df = counts.to_frame().reset_index()

            for col, domain in domains.items():
                assert col in category_cols, f"Unknown category column: {col}"
                other_cat = set(category_cols) - {col}

                count_df = (
                    count_df.groupby(list(other_cat), group_keys=True)
                    .apply(
                        lambda df: df.set_index(col)
                        .drop(columns=[*list(other_cat)])
                        .reindex(
                            domain,
                            fill_value=0,
                        )
                        .rename_axis(index=col)
                    )
                    .reset_index()
                )

            counts = count_df.set_index(category_cols)["value"]

    return counts



"""Utilities for working with common plots."""

from collections.abc import Callable, Mapping
from functools import reduce, wraps
from pathlib import Path
from textwrap import dedent, indent
from typing import Any, Concatenate, Literal, ParamSpec, TextIO, TypeAlias, cast
from uuid import uuid4

import pandas as pd
import plotly.graph_objects as go
from typing_extensions import deprecated

ImageFormat: TypeAlias = Literal["svg", "pdf", "png", "jpg", "webp"]


@deprecated("Use `plotly_to_html` instead.")
def export_html(
    fig: go.Figure,
    out: Path | str | TextIO,
    width: int = 800,
    height: int = 450,
    scale: float = 3,
    name: str = "plot",
    download_format: ImageFormat = "svg",
):
    """Save plotly figure to interactive html.

    Args:
        fig: Plotly figure to save.
        out: Path to save to or file-like object to write to.
        width: Width of the figure in pixels.
        height: Height of the figure in pixels.
        scale: Scale factor for the figure.
        name: Name of the figure.
        download_format: Format to use when downloading the figure.
    """
    fig = fig.update_layout(
        title=name,
    )

    with open(out, "w") if isinstance(out, Path | str) else out as f:
        f.write(
            plotly_to_html(
                fig=fig,
                download_scale=scale,
                download_format=download_format,
            )
        )


ResponsiveMode: TypeAlias = Literal[None, "keep-aspect", "stretch-width"]

default_responsiveness: dict[ResponsiveMode, int | None] = {
    None: 400,
    "keep-aspect": 720,
    "stretch-width": 1080,
}

with open(Path(__file__).parent / "responsive_plotly.js") as f:
    plotly_responsive_js: str = f.read()


def _ind(text: str, count: int) -> str:
    """Indent all lines in text except for first by count spaces."""
    return indent(text, " " * count).lstrip(" " * count)


def plotly_to_html(
    fig: go.Figure,
    write_to: Path | str | TextIO | None = None,
    responsive: bool = True,
    min_width: int = 500,
    max_width: int = 1500,
    max_height: int = 600,
    download_format: ImageFormat = "svg",
    download_scale: float = 3,
    full_html: bool = True,
    plotly_js_url: str
    | None = "https://cdn.jsdelivr.net/npm/plotly.js@2/dist/plotly.min.js",
) -> str:
    """Convert plotly figure to interactive (and responsive) html.

    Args:
        fig: Plotly figure to save.
        write_to: Path to save to or file-like object to write to.
        responsive: Whether to make the figure responsive.
        min_width: Minimum width of the figure in pixels.
        max_width: Maximum width of the figure in pixels.
        max_height: Maximum height of the figure in pixels.
        download_format: Format to use when downloading the figure.
        download_scale: Scale factor for the figure resolution when downloading.
        full_html: Whether to wrap the figure in a full html document.
        plotly_js_url:
            URL to load plotly.js library from.
            Defaults to the latest version on jsdelivr.
            Set to None to leave out the plotly.js script tag.

    Returns:
        HTML string.
    """
    # Get the name of the figure,
    # either from the figure itself or from the supplied argument.
    # Use fallback name "plot" if neither is available.
    name = (
        cast(dict, cast(go.Layout, fig.layout).to_plotly_json().get("title") or {}).get(
            "text"
        )
        or "plot"
    )

    fig_id = str(uuid4())[:10]
    script_id = str(uuid4())[:10]

    layout = cast(go.Layout, fig.layout).to_plotly_json()
    width = layout.get("width") or 720
    height = layout.get("height") or 480

    # Render the figure to html.
    fig_html: str = fig.to_html(
        include_plotlyjs=False,
        full_html=False,
        div_id=fig_id,
        config={
            "toImageButtonOptions": {
                "format": download_format,
                "filename": name,
                "height": height,
                "width": width,
                "scale": download_scale,
            },
        },
    )

    res_html = fig_html
    if responsive:
        # Define attributes to pass to the script tag.
        script_attrs = {
            "plotly-js-url": plotly_js_url,
            "fig-id": fig_id,
            "fig-width": width,
            "fig-height": height,
            "min-width": min_width,
            "max-width": max_width,
            "max-height": max_height,
        }

        script_attr_str = " ".join(
            f'{attr}="{value}"' for attr, value in script_attrs.items()
        )

        plotly_js_script = (
            dedent(
                f"""
            <script src="{plotly_js_url}"></script>
            """
            )
            if plotly_js_url is not None
            else ""
        )

        # Combine the figure html with a script tag that makes it responsive,
        # and optionally with the plotly.js script tag.
        res_html = dedent(
            f"""
            <div class="plotly-responsive-container" style="width: 100%;">
                {_ind(plotly_js_script, 16)}
                {_ind(fig_html, 16)}
                <script
                    id="{script_id}"
                    {_ind(script_attr_str, 20)}
                >
                    const scriptID = "{script_id}";
                    {_ind(plotly_responsive_js, 20)}
                </script>
            </div>
            """
        )

    # Wrap the figure in a full html document if requested.
    html = dedent(
        f"""
        <!doctype html>
        <html>
            <head>
                <title>{name}</title>
            </head>
            <body>
                {_ind(res_html, 16)}
            </body>
            </html>
         """
        if full_html
        else res_html
    )

    # Write the html to a file if requested.
    if write_to is not None:
        with open(write_to, "w") if isinstance(write_to, Path | str) else write_to as f:
            f.write(html)

    return html


P = ParamSpec("P")
PlottingFunction = Callable[Concatenate[pd.DataFrame, P], go.Figure]


def _merge_data(figs: Mapping[str, go.Figure | go.Frame]) -> list:
    return [d for fig in figs.values() for d in (fig.data or [])]


def _merge_layout(figs: Mapping[str, go.Figure | go.Frame]) -> dict[str, Any]:
    return reduce(
        lambda a, b: {**a, **b},
        [
            cast(dict, fig.layout.to_plotly_json())  # type: ignore
            for fig in figs.values()
        ],
    )


def _get_frames(figs: Mapping[str, go.Figure], label: str) -> dict[str, go.Frame]:
    return {
        group: frame
        for group, fig in figs.items()
        for f in fig.frames
        if (frame := cast(go.Frame, f)).name == label
    }


def with_dropdown(
    group_by: str,
    dropdown_kwars: dict[str, Any] | None = None,
) -> Callable[[PlottingFunction[P]], PlottingFunction[P]]:
    """Add a dropdown to a plotting function.

    Args:
        group_by: Column to group by for the dropdown.
        dropdown_kwars: Keyword arguments to pass to the dropdown layout.

    Returns:
        Decorator that adds a dropdown to the returned plot.
    """

    def decorator(func: PlottingFunction[P]) -> PlottingFunction[P]:
        """Add a dropdown to a plotting function."""

        @wraps(func)
        def wrapper(data: pd.DataFrame, *args: P.args, **kwargs: P.kwargs) -> go.Figure:
            figs = {
                str(group): func(sub_data, *args, **kwargs)
                for group, sub_data in data.groupby(group_by)
            }

            fig_data = _merge_data(figs)
            layout = _merge_layout(figs)

            frames = None
            if "sliders" in layout:
                labels = [step["label"] for step in layout["sliders"][0]["steps"]]
                frames = []
                for label in labels:
                    sub_frames = _get_frames(figs, label)
                    frames.append(
                        go.Frame(
                            data=_merge_data(sub_frames),
                            layout=_merge_layout(sub_frames),
                            name=label,
                        )
                    )

            split_fig = go.Figure(
                data=fig_data,
                layout=layout,
                frames=frames,
            )

            trace_sources = pd.Series(
                [group for group, fig in figs.items() for _ in fig.data]
            )

            split_fig.update_layout(
                updatemenus=[
                    *(
                        [split_fig.layout["updatemenus"][0]]  # type: ignore
                        if "updatemenus" in layout
                        else []
                    ),
                    {
                        "active": 0,
                        "buttons": [
                            {
                                "label": group,
                                "method": "update",
                                "args": [
                                    {
                                        "visible": (
                                            trace_sources == str(group)
                                        ).to_list(),
                                    },
                                ],
                            }
                            for group in figs.keys()
                        ],
                        "direction": "up",
                        "pad": {"r": 10, "t": 70},
                        "showactive": True,
                        "x": 1.1,
                        "xanchor": "left",
                        "y": 0,
                        "yanchor": "top",
                        **(dropdown_kwars or {}),
                    },
                ]
            )

            first_group = list(figs.keys())[0]
            split_fig.for_each_trace(
                lambda trace: trace.update(visible=trace.name == first_group)
            )

            return split_fig

        return wrapper

    return decorator



"""Test reflect module."""

import py_research
from py_research.reflect import PyObjectRef, env_info, get_outdated_deps


class StaticObject:
    """Static object for test."""


def test_py_obj_ref():
    """Test Python object referencing."""
    ref = PyObjectRef.reference(StaticObject)
    assert ref.package == "py-research"

    obj = ref.resolve()
    assert obj is StaticObject


def test_get_outdated_deps():
    """Test get_outdated_deps function."""
    outdated_deps = get_outdated_deps(py_research)
    assert isinstance(outdated_deps, dict)


def test_env_info():
    """Test env_info function."""
    info = env_info()
    assert isinstance(info, dict)
    assert "repo" in info
    assert "requirements" in info
    assert "python_version" in info



"""Test plots module."""

from pathlib import Path
from tempfile import gettempdir

import pandas as pd
import plotly.express as px
import plotly.graph_objects as go

from py_research.plots import export_html, plotly_to_html, with_dropdown


def plot_func(df: pd.DataFrame) -> go.Figure:
    """Plot a scatter plot of GDP per capita vs population with a slider for year."""
    return px.scatter(
        df,
        x="pop",
        y="gdpPercap",
        color="country",
        size="lifeExp",
        animation_frame="year",
    )


def test_export_html():
    """Test export_html."""
    # Create a simple figure
    fig = plot_func(df=px.data.gapminder())

    # Test with a Path object
    path = Path(gettempdir()) / "test_plot_1.html"
    try:
        export_html(fig, path)
        assert path.is_file()
    finally:
        path.unlink()  # Clean up


def test_plotly_to_html():
    """Test plotly_to_html."""
    path = Path(gettempdir()) / "test_plot_2.html"
    # Create a simple figure
    fig = plot_func(df=px.data.gapminder())

    html = plotly_to_html(fig, write_to=path)

    assert isinstance(html, str)
    assert path.is_file()


def test_with_dropdown():
    """Test ``with_dropdown`` decorator.""" ""
    df = px.data.gapminder()
    plot_func_wd = with_dropdown("continent")(plot_func)

    fig = plot_func_wd(df)
    assert isinstance(fig, go.Figure)
